from typing import Optional

from starlette.requests import Request

IP_HEADER = 'X-Forwarded-For'


def get_ip(request: Request) -> Optional[str]:
    ips = request.headers.get(IP_HEADER)
    if ips:
        return ips.split(',', 1)[0].strip(' ')
    elif client := request.scope.get('client'):
        return client[0]



import asyncio
import os
import sys
from dataclasses import dataclass
from typing import List

from aiohttp import web
from aiohttp.test_utils import TestServer
from aiohttp.web import Application
from aiohttp.web_exceptions import HTTPException
from aiohttp.web_middlewares import middleware
from aiohttp.web_response import Response, json_response
from async_timeout import timeout

__all__ = 'create_dummy_server', 'DummyServer', 'Offline'


async def return_any_status(request):
    status = int(request.match_info['status'])
    # TODO how do we deal with 301 extra which should have the "Location" header
    return Response(text=f'test response with status {status}', status=status)


async def recaptcha_dummy(request):
    data = await request.post()
    response = data.get('response')
    request['log_msg'] = f'recaptcha {response}'
    if response == '__ok__':
        return json_response(dict(success=True, hostname='testserver'))
    elif response == '__wrong_host__':
        return json_response(dict(success=True, hostname='__wrong_host__'))
    elif response == '__400__':
        return json_response({}, status=400)
    elif response and response.startswith('__ok__ host:'):
        return json_response(dict(success=True, hostname=response.split(':', 1)[-1]))
    else:
        return json_response(dict(success=False, hostname='testserver'))


@middleware
async def log_middleware(request, handler):
    try:
        r = await handler(request)
    except Exception as e:
        status = e.status if isinstance(e, HTTPException) else 500
        request.app['log'].append(f'{request.method} {request.path_qs} > {status}')
        raise
    else:
        log = f'{request.method} {request.path_qs} > {r.status}'
        if 'log_msg' in request:
            log += ' ({})'.format(request['log_msg'])
        request.app['log'].append(log)
        return r


def create_dummy_app() -> Application:
    app = web.Application(middlewares=(log_middleware,))
    app.add_routes(
        [
            web.route('*', r'/status/{status:\d+}/', return_any_status, name='any-status'),
            web.post('/recaptcha_url/', recaptcha_dummy, name='recaptcha-dummy'),
        ]
    )
    app['log'] = []
    return app


@dataclass
class DummyServer:
    server: TestServer
    app: Application
    log: List
    server_name: str

    async def stop(self):
        await self.server.close()


async def create_dummy_server(loop, *, extra_routes=None, extra_context=None) -> DummyServer:
    """
    Setup and start a dummy server to accept HTTP requests for testing
    """
    app = create_dummy_app()
    if extra_routes:
        app.add_routes(extra_routes)
    if extra_context:
        app.update(extra_context)

    server = TestServer(app)
    await server.start_server(loop=loop)

    return DummyServer(server, app, app['log'], f'http://localhost:{server.port}')


class Offline:
    """
    Lazy pytest decorator to skip tests if you're currently not online.

    Usage:
        _offline = Offline()
        skip_if_offline = pytest.mark.skipif(_offline, reason='not online')

        @skip_if_offline
        def test_whatever():
            ...
    """

    def __init__(self, loop=None):
        self._loop = loop
        self.is_offline = None

    def __bool__(self):
        if self.is_offline is None:
            loop = self._loop or asyncio.new_event_loop()
            self.is_offline = loop.run_until_complete(self._check())
        return self.is_offline

    async def _check(self):
        if os.getenv('CI'):
            # on CI we should always be online
            # I assume that all CI services set the CI environment variable!
            return False

        import aiodns

        resolver = aiodns.DNSResolver()
        try:
            with timeout(1):
                await resolver.query('google.com', 'A')
        except (aiodns.error.DNSError, asyncio.TimeoutError) as e:
            print(f'\nnot online: {e.__class__.__name__} {e}\n', file=sys.stderr)
            return True
        else:
            return False



from typing import TYPE_CHECKING

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request

__all__ = 'PgMiddleware', 'get_db'

if TYPE_CHECKING:
    from buildpg.asyncpg import BuildPgConnection
    from starlette.responses import Response

    from ..middleware import CallNext


class GetPgConn:
    __slots__ = '_glove', '_conn'

    def __init__(self, glove):
        self._glove = glove
        self._conn = None

    async def __call__(self):
        if self._conn is None:
            self._conn = await self._glove.pg.acquire()
        return self._conn

    async def release(self):
        if self._conn is not None:
            conn = self._conn
            self._conn = None
            await self._glove.pg.release(conn)


class PgMiddleware(BaseHTTPMiddleware):
    def __init__(self, app):
        super().__init__(app)

        from ..main import glove

        self.glove = glove

    async def dispatch(self, request: Request, call_next: 'CallNext') -> 'Response':
        request.state.get_pg_conn = GetPgConn(self.glove)
        try:
            return await call_next(request)
        finally:
            await request.state.get_pg_conn.release()


async def get_db(request: Request) -> 'BuildPgConnection':
    return await request.state.get_pg_conn()



import asyncio
import logging
import re
from dataclasses import dataclass
from enum import Enum
from importlib import import_module
from typing import Callable, Dict, Type

from .. import glove
from .helpers import DummyPgPool

logger = logging.getLogger('foxglove.patch')
patches = []
__all__ = 'run_patch', 'patch', 'update_enums', 'run_sql_section'


@dataclass
class Patch:
    func: Callable
    direct: bool = False


def run_patch(patch_name: str, live: bool, args: Dict[str, str]):
    for path in getattr(glove.settings, 'patch_paths', []):
        import_module(path)

    if patch_name is None:
        logger.info(
            'available patches:\n{}'.format(
                '\n'.join('  {}: {}'.format(p.func.__name__, (p.func.__doc__ or '').strip('\n ')) for p in patches)
            )
        )
        return 0

    patch_lookup = {p.func.__name__: p for p in patches}
    try:
        patch = patch_lookup[patch_name]
    except KeyError:
        logger.error('patch "%s" not found in patches: %s', patch_name, [p.func.__name__ for p in patches])
        return 1

    if patch.direct:
        if not live:
            logger.error('direct patches must be called with "--live"')
            return 1
        logger.info(f'running patch {patch_name} direct')
    else:
        logger.info(f'running patch {patch_name} live {live}')
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(_run_patch(patch, live, args)) or 0


async def _run_patch(patch: Patch, live: bool, args: Dict[str, str]):
    from .main import lenient_conn

    conn = await lenient_conn(glove.settings)
    tr = None
    if not patch.direct:
        tr = conn.transaction()
        await tr.start()
    logger.info('=' * 40)
    glove.pg = DummyPgPool(conn)
    await glove.startup()
    kwargs = dict(conn=conn, live=live, args=args, logger=logger)
    try:
        if asyncio.iscoroutinefunction(patch.func):
            result = await patch.func(**kwargs)
        else:
            result = patch.func(**kwargs)
        if result is not None:
            logger.info('result: %s', result)
    except BaseException:
        logger.info('=' * 40)
        logger.exception('Error running %s patch', patch.func.__name__)
        if not patch.direct:
            await tr.rollback()
        return 1
    else:
        logger.info('=' * 40)
        if patch.direct:
            logger.info('committed patch')
        else:
            if live:
                logger.info('live, committed patch')
                await tr.commit()
            else:
                logger.info('not live, rolling back')
                await tr.rollback()
    finally:
        await glove.shutdown()
        await conn.close()


def patch(func_=None, /, direct=False):
    if func_:
        patches.append(Patch(func=func_))
        return func_
    else:

        def wrapper(func):
            patches.append(Patch(func=func, direct=direct))
            return func

        return wrapper


@patch
async def rerun_sql(*, conn, **kwargs):
    """
    rerun the contents of settings.sql_path.
    """
    # this require you to use "CREATE X IF NOT EXISTS" everywhere
    await conn.execute(glove.settings.sql)


async def update_enums(enums: Dict[str, Type[Enum]], conn):
    """
    update sql enums from python enums, this requires @patch(direct=True) on the patch
    """
    for name, enum in enums.items():
        for t in enum:
            await conn.execute(f"ALTER TYPE {name} ADD VALUE IF NOT EXISTS '{t.value}'")


async def run_sql_section(section_name, sql, conn):
    """
    Run a section of a sql string (eg. settings.sql) based on tags in the following format:
        -- { <chunk name>
        <sql to run>
        -- } <chunk name>
    """
    m = re.search(f'^-- *{{+ *{section_name}(.*)^-- *}}+ *{section_name}', sql, flags=re.DOTALL | re.MULTILINE)
    if not m:
        raise RuntimeError(f'chunk with name "{section_name}" not found')
    logger.info('run_sql_section running section "%s"', section_name)
    sql = m.group(1).strip(' \n')
    await conn.execute(sql)



from typing import Callable

from fastapi import Request, Response
from fastapi.routing import APIRoute

__all__ = ('KeepBodyAPIRoute',)


class KeepBodyRequest(Request):
    async def body(self) -> bytes:
        if not hasattr(self, '_body'):
            chunks = []
            async for chunk in self.stream():
                chunks.append(chunk)
            self.scope['_body'] = self._body = b''.join(chunks)
        return self._body


class KeepBodyAPIRoute(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            request = KeepBodyRequest(request.scope, request.receive)
            return await original_route_handler(request)

        return custom_route_handler



import asyncio
from typing import Optional

from buildpg.asyncpg import BuildPgConnection


class TimedLock(asyncio.Lock):
    def __init__(self, *, timeout=0.5):
        self.timeout = timeout
        super().__init__()

    async def acquire(self):
        try:
            await asyncio.wait_for(super().acquire(), timeout=self.timeout)
        except asyncio.TimeoutError as e:
            raise asyncio.TimeoutError('DummyPg query lock timed out') from e


class _LockedExecute:
    def __init__(self, conn: BuildPgConnection, lock: Optional[TimedLock] = None):
        self._conn: BuildPgConnection = conn
        # could also add lock to each method of the returned connection
        self._lock: TimedLock = lock or TimedLock()

    async def execute(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.execute(*args, **kwargs)

    async def execute_b(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.execute_b(*args, **kwargs)

    async def fetch(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.fetch(*args, **kwargs)

    async def fetch_b(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.fetch_b(*args, **kwargs)

    async def fetchval(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.fetchval(*args, **kwargs)

    async def fetchval_b(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.fetchval_b(*args, **kwargs)

    async def fetchrow(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.fetchrow(*args, **kwargs)

    async def fetchrow_b(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.fetchrow_b(*args, **kwargs)

    async def executemany(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.executemany(*args, **kwargs)

    async def executemany_b(self, *args, **kwargs):
        async with self._lock:
            return await self._conn.executemany_b(*args, **kwargs)


class DummyPgTransaction(_LockedExecute):
    _tr = None

    async def __aenter__(self):
        async with self._lock:
            self._tr = self._conn.transaction()
            await self._tr.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        async with self._lock:
            if exc_type:
                await self._tr.rollback()
            else:
                await self._tr.commit()
            self._tr = None


class DummyPgConn(_LockedExecute):
    def transaction(self):
        return DummyPgTransaction(self._conn, self._lock)

    def __repr__(self) -> str:
        return f'<DummyPgConn {self._conn._addr} {self._conn._params}>'


class _ConnAcquire:
    def __init__(self, conn: BuildPgConnection, lock: TimedLock):
        self._conn = conn
        self._lock = lock

    async def __aenter__(self) -> DummyPgConn:
        return DummyPgConn(self._conn, self._lock)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def _get_conn(self) -> DummyPgConn:
        return DummyPgConn(self._conn, self._lock)

    def __await__(self):
        return self._get_conn().__await__()


class DummyPgPool(_LockedExecute):
    """
    dummy connection pool useful for testing, only one connection is used, but this will behave like
    Connection or BuildPgConnection, including locking before using the underlying connection.
    """

    def acquire(self):
        return _ConnAcquire(self._conn, self._lock)

    async def close(self):
        pass

    async def release(self, conn):
        pass

    def as_dummy_conn(self) -> DummyPgConn:
        """
        convert to a DummyPgConn.

        **THIS IS OBVIOUSLY ONLY TO BE USED IN TESTS**
        """
        return DummyPgConn(self._conn, self._lock)

    def __repr__(self) -> str:
        return f'<DummyPgPool {self._conn._addr} {self._conn._params}>'


class SyncDb:
    def __init__(self, conn: BuildPgConnection, loop: asyncio.AbstractEventLoop):
        self._conn = conn
        self._loop = loop

    def execute(self, *args, **kwargs):
        return self._loop.run_until_complete(self._conn.execute(*args, **kwargs))

    def execute_b(self, *args, **kwargs):
        return self._loop.run_until_complete(self._conn.execute_b(*args, **kwargs))

    def fetch(self, *args, **kwargs):
        v = self._loop.run_until_complete(self._conn.fetch(*args, **kwargs))
        return [dict(r) for r in v]

    def fetch_b(self, *args, **kwargs):
        v = self._loop.run_until_complete(self._conn.fetch_b(*args, **kwargs))
        return [dict(r) for r in v]

    def fetchval(self, *args, **kwargs):
        return self._loop.run_until_complete(self._conn.fetchval(*args, **kwargs))

    def fetchval_b(self, *args, **kwargs):
        return self._loop.run_until_complete(self._conn.fetchval_b(*args, **kwargs))

    def fetchrow(self, *args, **kwargs):
        v = self._loop.run_until_complete(self._conn.fetchrow(*args, **kwargs))
        return None if v is None else dict(v)

    def fetchrow_b(self, *args, **kwargs):
        v = self._loop.run_until_complete(self._conn.fetchrow_b(*args, **kwargs))
        return None if v is None else dict(v)

    def executemany(self, *args, **kwargs):
        return self._loop.run_until_complete(self._conn.executemany(*args, **kwargs))

    def executemany_b(self, *args, **kwargs):
        return self._loop.run_until_complete(self._conn.executemany_b(*args, **kwargs))
